;PROGRAM:  BigBuffer
;BY:       Jonathan Kraidin 1.0 on 8/31/90
;PURPOSE:  To extend keyboard buffer past the 16 limit
;OVERVIEW: The INT9 is serviced and the buffer is examined.  If a keystroke
;          has been added it is removed and placed into the extended buffer.
;          INT8 waits until the keyboard buffer is empty; when it is, it
;          stuffs the first item from its buffer into the keyboard buffer.  
;          This process repeats each time INT8 is called until the extended
;          buffer is empty.
;USE: Use freely, no copyright.
; revision: 1.10  - cleared contents of keyboard buffer as Extender drains
;                   it so that if other programs, such as a keyboard rate
;                   enhancer, that is remembering the previous key and watching
;                   for Tail movement, will not think the same key is pressed
;                   when we move the Tail back (even during a key release!).
;                   If we did not do this, the possibility exists that when
;                   we move the Tail past the keys the enhancer put in, 
;                   the Tail points to a location in the buffer that contains
;                   the previous key pressed because the buffer contents were
;                   not erased when WE drained it. The enhancer thinks that
;                   the same key is being held down and continues to stuff 
;                   the keyboard buffer.

Bios_data Segment at 40h
     org 1Ah
     Head     DW ?               ; first in location
     Tail     DW ?               ; last in location
     Buffer   DW 16 Dup (?)      ; data itself
     Buffer_End Label Word       ; end of buffer marker
Bios_data ENDS

cseg segment public word 'code'
     assume cs:cseg, ds:nothing, es:nothing, ss:nothing
     org 100h
     entry: jmp init

    Speaker_port EQU   61h
    Num_cycles   EQU   20              ; length of tone
    Half_Cycle2  EQU   200             ; length of pulse (waveform)
				       ;  i.e., period between Push and Pulls!
    Old_Int9 Label Dword     
	   Old_Int9_off DW ?
	   Old_Int9_seg DW ?
    Old_Int8 Label Dword
	   Old_Int8_off DW ?
	   Old_Int8_seg DW ?
    Buff_Size EQU 60
    MyBuffer DW Buff_Size DUP (?)      ; our extended buffer
    Entry_Count    DW 0                ; number of entries in our buffer
    Last_Entry     DW 0                ; location of our entry into buffer
    Drain_Start    DW ?                ; location of beginning and end of
    Drain_End      DW ?                ;  the data we move to our ext. buffer
    Last_Tail      DW ?                ; last item we stuffed into key. buffer
    Intro_Message  DB 'Keyboard buffer extender 1.10 by Jonathan Kraidin',10,13
		   DB 'Type ahead buffer set to between 44 and 61.',10,13,'$'

; **************************** Begin Code **************************************
;------------------------------------------------------------------------------
; Beep2 - driver for producing a speaker tone
; INPUT:  NOTHING
; OUTPUT: NOTHING
; REGS:   ALL PRESERVED
;------------------------------------------------------------------------------
Beep2 proc near
	 assume ds:nothing
	 push bx                      ; save some registers
	 push dx
	 push cx
	 mov  cx, 0FFFh
   delay_loop:
	 inc dx
	 loop delay_loop
	 mov dx, half_cycle2          ; get half_cycle length
	 mov bx, num_cycles           ; get length of tone
	 call Tone
	 pop  cx
	 pop  dx
	 pop  bx
	 ret
Beep2 endp

;------------------------------------------------------------------------------
;Tone -- produces speaker tone
;INPUT:  DX: length of half cycle
;        CX: number of cycles for tone
;OUTPUT: NONE
;REGS:   ALL PRESERVED
;------------------------------------------------------------------------------
Tone proc near
	assume ds:nothing
	push ax                    ; save some registers
	push bx
	push cx

	in   al, Speaker_Port      ; get original port value
	mov  ah, al                ; save it for end of routine
 DoTone:
	and  al, 11111110b         ; disconnect from timer chip
 Start_Tone:
	or   al, 11111110b         ; turn on speaker (bit 1 = 1)
	out  Speaker_Port, al      ; push!
	mov  cx, dx                ; get length of half cycle
 ToneLoop1:
	loop ToneLoop1
	 
	and  al, 11111100b         ; turn speaker off (bit 1 = 0)
	out  Speaker_Port, al      ; pull!
	mov  cx, dx
 ToneLoop2:
	loop ToneLoop2
	dec  bx
	jnz  Start_Tone            ; execute a number of cycles
	
	mov  al, ah                ; get original value at port
	out  Speaker_Port, al  
 
	pop  cx
	pop  bx
	pop  ax
	ret
Tone endp

;----------------------------- debug procedure ----------------------------
display2 proc near
   assume ds:nothing
   push es
   push ax
   push bx
   mov  ax, 0b800h
   mov  es, ax
   mov  bx, entry_count
   shl  bx, 1
   mov  es:[(80*20)+bx], cx
   pop  bx
   pop  ax
   pop  es
   ret
display2 endp
;-----------------------------------------------------------------------------
;------------------------------------------------------------------------------
;PROCEDURE: Our_Int9
;PURPOSE:   Watch for new keystrokes in the keyboard buffer
;INPUT:     NOTHING
;OUTPUT:    Drains keyboard buffer and places data into extended buffer
;REGS:      ALL PRESERVED
;IMORTANT POINTS:  In general, all keystrokes after the first one in the
;           buffer are transfered to the extended buffer.  This allows DOS to
;           process the lone keystroke.  When the buffer is empty, INT8 knows
;           to put any extended buffer data back into the keyboard buffer one
;           item at a time.  If all keyboard data were transfered then when
;           INT8 stuffs the keyboard with a keystroke, INT9 would remove it 
;           again.  Solutions are possible, but why bother?  One must be wary
;           of INT8 stuffing the keyboard buffer while the user is typing new
;           keys.  We solve this by remembing the last item stuffed in the buff:
;           Last_Entry.  If we stuff the buffer and a key is typed after it, 
;           INT9 removes it (all keys after the first buffer entry).  If we
;           stuff the buffer at the head, the user types a keystroke, and DOS
;           processes the entry at the Head, we see that the new keystroke at
;           the new head is not the same as Last_Entry. If our ext. buffer has
;           data we know to remove everything from the keyboard buffer and
;           make everyone wait in the processing line properly.
; .call INT9 handler
; .check if extended buffer full
; .check if keyboard buffer
; .check if user typing as extended buffer is dumping to keyboard buffer
; .move extended buffer data to keyboard buffer
; .cleanup and exit
;------------------------------------------------------------------------------
Our_Int9 PROC NEAR
   push ds
   push es
   push ax
   push bx
   push cx
   pushf

   pushf                              ; simulate interrupt
   call Old_Int9                      ;  to see what we bring back
   
   assume ds:bios_data                ; set bios data segment
   mov  ax, bios_data
   mov  ds, ax

   cmp  Entry_count, (Buff_Size - 16) ; is our ext. buffer nearly full?
   jb   resume                        ; NO -->
   call Beep2                         ; YUP. Let user know it!
   mov  bx, Last_Tail
   mov  tail, bx                      ; discard recent entries to key. buffer
   jz   Exit_Int9                     ; out'a'here, dude.

 resume:
   cli                              ; no interruptions, please.
; ---------- keyboard buffer begins at head and grows along the tail -------
   mov  bx, tail                    ; get end of keyboard buffer
   cmp  bx, head                    ; compare it with the head
   jz   Exit_Int9                   ; if equal buffer empty --> exit

; ---------- keyboard has data ---------------------------------------------
   cmp  Last_Entry, 0               ; did we stuff buffer last time?
   jz   Case1                       ; NO --> 
   mov  ax, head                    ; get first entry location
   cmp  ax, Last_Entry              ; are they the same
   jz   Case1                       ; YES --> head did not move and therefore
				    ;  a new keystroke did not bump ahead of
				    ;  the entries in our extended buffer
				    ; NO:  thus, every keystroke in buffer
				    ;  is new and must come after the earlier
				    ;  entries in the extended buffer
   mov  bx, head                    ; looks like we move all entries
   jmp  Case2                       ; so lets continue at Case2
 Case1:
   sub  bx, 2                       ; Is there just one entry in buffer?
   call check_boundary              ;   adjust tail-2 for buffer wrap
   cmp  bx, head                    ;   is this equal to head?
   jz   Exit_Int9                   ; YES --> exit (just one entry)
; ---------- move data to our buffer ----------------------------------------
   mov  bx, head                                     
   add  bx, 2                       ; skip first entry
   call check_boundary
 Case2:
   mov  drain_start, bx
   mov  bx, tail                    ; set the end of data to move
   mov  drain_end, bx
   call drain_buffer                ; drain keyboard buffer to our ext. buffer
   mov  bx, drain_start             ; cleanup keyboard buffer
   mov  tail, bx                    ;  by reseting tail
   mov  last_tail, bx               ; remember last location of tail
 Exit_Int9:
   sti                              ; tidy up.
   popf
   pop  cx
   pop  bx
   pop  ax
   pop  es
   pop  ds
   iret
Our_Int9 ENDP

;----------------------------------------------------------------------------
;PROCEDURE: Our_Int8
;PURPOSE:   To stuff extended keyboard buffer data back into keyboard buffer
;           when the keyboard buffer is empty
;INPUT:     NOTHING
;OUTPUT:    Moves data into keyboard buffer and removes it from extended buffer
;REGS:      ALL PRESERVED
; .process INT8
; .check if items in our buffer
; .check if keyboard buffer empty and ready to receive next keystroke
; .stuff keyboard buffer and adjust extended buffer
; .reset keyboard tail
;------------------------------------------------------------------------------
Our_Int8 PROC NEAR
    assume ds:bios_data
    push ds
    push ax
    push bx
    push cx
    pushf

    pushf                           ; simulate interrupt
    call Old_Int8

    mov  ax, bios_data
    mov  ds, ax
    
    cmp  Entry_Count, 0             ; anything in our extended buffer?
    jnz  Continue                   ; YES --> continue
    xor  ax, ax
    mov  Last_Entry, ax             ; no more entries in our ext. buffer
    jmp  Exit_Int8                  ;   to worry about
    
 continue:
    cli                             ; no interruptions
    mov  bx, tail
    cmp  bx, head                   ; is keyboard buffer empty?
    jnz  Exit_Int8                  ; NO --> exit
    call Slide_MyBuff               ; get next data from our ext. buffer in BX
    mov  ax, bx                     ; make a copy of it
    mov  bx, head                   
    mov  [bx], ax                   ; stuff keyboard buffer with entry
    ;mov  cx, ax
    ;call display2
    mov  Last_Entry, bx             ; remember what we just put in
    mov  bx, tail                   ; prepare to update tail 
    add  bx, 2
    call check_boundary             ; check for wrap
    mov  tail, bx                   ; adjust tail
    mov  last_tail, bx              ; remember this, too
 Exit_Int8:
    sti
    popf
    pop  cx
    pop  bx
    pop  ax
    pop  ds
    iret
Our_Int8 ENDP

;----------------------------------------------------------------------------
;PROCEDURE: Drain_Buffer
;PURPOSE:   To move data from keyboard buffer into our extended buffer
;INPUT:     Drain_Start (S), Drain_End (E) = segment to move
;           S       E     (S points to first data, E points one beyond last)
;           d a t a           
;
;OUTPUT:    Data moved to Extended Buffer
;           Keyboard Tail reset
;-----------------------------------------------------------------------------
Drain_Buffer PROC NEAR
    assume ds:nothing
    push es
    push ax
    push bx
    push di

    assume es:bios_data
    mov  ax, bios_data                ; get bios segment
    mov  es, ax

    mov  bx, Drain_start              ; set start of data we will move
    mov  di, Entry_Count              ; set location in our buffer to move to
    shl  di, 1                        ;  and convert to a Word offset
    

  Drain_Loop:
    mov  ax, es:[bx]                  ; get data from keyboard buffer
    mov  MyBuffer[di], ax             ; and put in our buffer
    xor  ax, ax
    mov  es:[bx], ax
    add  bx, 2                        ; move to next item in buffer
    call check_boundary
    add  di, 2
    inc  Entry_Count                  ; keep track of entries in our buffer
    cmp  bx, Drain_End                ; is this the end of our data?
    jne  Drain_loop                   ; if equal/above we are done

stopit:
    pop  di
    pop  bx
    pop  ax
    pop  es
    ret
Drain_Buffer ENDP

;---------------------------------------------------------------------------
;PROCEDURE: Check_Boundary
;PURPOSE:   To adjust Tail within buffer boundaries
;INPUT:     BX = Tail
;OUTPUT:    BX = adjusted Tail
;---------------------------------------------------------------------------
Check_Boundary PROC NEAR
    assume ds:bios_data
    push ds
    push ax
    mov  ax, bios_data               ; get right data segment
    mov  ds, ax
  check_high:
    cmp  bx, offset (buffer_end - 2) ; is tail past end?
    jbe  check_low                   ; no -->
    mov  bx, offset buffer           ; set tail to beginning
    jmp  exit_this                   ; no -->
  check_low:
    cmp  bx, offset buffer           ; is tail past beginning?
    jae  exit_this                   ; no -->
    mov  bx, offset (buffer_end - 2) ; set to end
  exit_this:
    pop  ax
    pop  ds
    ret
Check_Boundary ENDP

;-------------------------------------------------------------------------
;PROCEDURE: Slide_MyBuff
;PURPOSE:   To slide our buffer down one notch
;INPUT:     Entry_Count = Entries in our extended buffer
;OUTPUT:    BX = data to enter into keyboard buffer
;           Entry_Count decremented
;-------------------------------------------------------------------------
Slide_MyBuff PROC NEAR
    assume ds:nothing
    push ax
    push cx
    push si
    mov  cx, Entry_Count            ; get number of entries
    mov  bx, MyBuffer               ; get first data
    xor  si, si
slide_loop:
    mov  ax, MyBuffer[si+2]
    mov  MyBuffer[si], ax           ; slide data
    add  si, 2
    loop slide_loop                 ; do this for all entries
    xor  ax, ax
    mov  MyBuffer[si], ax           ; zero last entry
    dec  Entry_Count                ; adjust entry count
    pop  si
    pop  cx
    pop  ax
    ret
Slide_MyBuff ENDP

;==================== TSR CURTAIN ===========================================
init:
	assume ds:nothing
	;--------------- chain another handler ----------
	mov  ax, cs
	mov  ds, ax
	mov  ax, 3509h                    ; set our INT9 handler
	int  21h
	mov  old_int9_off, bx
	mov  old_int9_seg, es
	mov  ax, 2509h
	lea  dx, Our_Int9
	int  21h
	mov  ax, 3508h                    ; set our INT8 handler
	int  21h
	mov  old_int8_off, bx
	mov  old_int8_seg, es
	mov  ax, 2508h
	lea  dx, Our_int8
	int  21h
	mov  dx, offset Intro_Message     ; display message    
	mov  ah, 9
	int  21h
	mov  ax, 3100h                    ; TSR time!
	mov  dx, (init - entry + 100h + 15) shr 4
	int  21h
cseg ends
     end  entry

ax, 3100h                    ; TSR time!
	mov  dx, (init - entry + 100h + 15) shr 4
	int  21h
cseg ends
     end  entry
